﻿// -----------------------------------------------------------------------
// <copyright file="MsSqlConversionHelper.cs" company="CodePlex">
//   Copyright © CodePlex 2012
// </copyright>
// -----------------------------------------------------------------------

namespace DataFrameworkGenerator.Helpers
{
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Data.OleDb;
    using System.Diagnostics;
    using System.Linq;

    using DataFrameworkGenerator.Factories;
    using DataFrameworkGenerator.Interfaces;

    /// <summary>
    /// This helpler assists with converting a Microsoft SQL Server schema into objects that can be built by the generator.
    /// </summary>
    public class MsSqlConversionHelper : IConversionHelper
    {
        /// <summary>
        /// The type we use for all parameters that need the database type specified.
        /// </summary>
        private const DatabaseType DatabaseTypeForHelper = DatabaseType.MsSql;

        /// <summary>
        /// Gets or sets the target schema.
        /// </summary>
        /// <value>The target schema.</value>
        public string TargetSchema { get; set; }

        /// <summary>
        /// Converts the type of the data type int to db data.
        /// </summary>
        /// <param name="metadata">The metadata.</param>
        /// <returns>a propertly formated DbType name.</returns>
        public string ConvertDataTypeToDbDataType(ColumnMetadata metadata)
        {
            // NOTE: http://social.msdn.microsoft.com/Forums/da/adodotnetdataproviders/thread/3fa2bc10-8720-49db-9d5d-0fa9554b65fe
            string convertedType;
            switch (metadata.ProviderDataType)
            {
                case 2: // Confirmed. =======================
                    convertedType = "DbType.Int16";
                    break;
                case 3: // Confirmed. ===========================
                    convertedType = "DbType.Int32";
                    break; 
                case 4: // Real. Confirmed ===========================
                    convertedType = "DbType.Single"; 
                    break;
                case 5: // Float. Confirmed ==================================
                    convertedType = "DbType.Single";
                    break;
                case 6: // Money. Confimred =================================
                    convertedType = "DbType.Currency";
                    break;
                case 11: // Confirmed. ===================================
                    convertedType = "DbType.Boolean";
                    break;
                case 12: // Time. Confirmed. ====================================
                    convertedType = "DbType.Time";
                    break;
                case 17: // Confirmed. =========================================
                    convertedType = "DbType.Byte"; 
                    break;
                case 20: // Confirmed. ====================================
                    convertedType = "DbType.Int64"; 
                    break;
                case 72: // Confirmed. =======================================
                    convertedType = "DbType.Guid";
                    break;
                case 128: // TimeStamp. Confirmed. ====================================
                    convertedType = "DbType.Binary"; 
                    break;
                case 129: // Confirmed. ====================================
                    convertedType = "DbType.StringFixedLength";
                    break;
                case 130: // NChar. Confirmed. ===================================
                    convertedType = "DbType.String";
                    break;
                case 131: // Confirmed. =============================================
                    convertedType = "DbType.Decimal";
                    break;
                case 132: // SQL Geography / Geometry / HierarchyId type
                    throw new NotSupportedException("The Geography, Geometry and HierarchyId types are not supported.");
                case 133: // Confirmed. =================================
                    convertedType = "DbType.Date";
                    break;
                case 135: // Confirmed. ==============================
                    convertedType = "DbType.DateTime"; // DbType.DateTime or DbType.DateTime2
                    break;
                case 200: // Confirmed. =========================
                    convertedType = "DbType.String";
                    break;
                case 201: // Confirmed. =========================
                    convertedType = "DbType.String";
                    break;
                case 202: // XML / String. Confirmed. =========================
                    convertedType = "DbType.String";
                    break;
                case 203: // NText. Confirmed. =========================
                    convertedType = "DbType.String";
                    break;
                case 204: // TimeStamp. Confirmed. ====================================
                    convertedType = "DbType.Binary";
                    break;
                case 205: // Image. Confirmed. ====================================
                    convertedType = "DbType.Binary";
                    break;
                default:
                    throw new InvalidOperationException(string.Format("Can not convert provided DB type \"{0}\" to a DbType.", metadata.DataType));
            }

            return convertedType;
        }

        /// <summary>
        /// Converts the type of the data type int to code.
        /// </summary>
        /// <param name="metadata">The metadata.</param>
        /// <returns>a properly formatted c# variable type.</returns>
        public string ConvertDataTypeToCodeType(ColumnMetadata metadata)
        {
            return metadata.DataType.Name;
        }

        /// <summary>
        /// Fetches the db tables.
        /// </summary>
        /// <param name="connectionString">The connection string.</param>
        /// <returns>
        /// A list containing all the tables in the provided schema.
        /// </returns>
        public IList<string> FetchDbTables(string connectionString)
        {
            List<string> tables = null;
            var connection = ConnectionFactory.GenerateConnection(DatabaseTypeForHelper, connectionString);

            var conn = connection as OleDbConnection;
            if (conn != null)
            {
                using (conn)
                {
                    conn.Open();

                    // Restrictions =  {TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, TABLE_TYPE}
                    // TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, TABLE_TYPE, TABLE_GUID, DESCRIPTION, TABLE_PROPID, DATE_CREATED, DATE_MODIFIED
                    var tblRestrictions = new[] { null, this.TargetSchema, null, "TABLE" };
                    var schemaTables = conn.GetSchema("tables", tblRestrictions);
                    tables = (from DataRow t in schemaTables.Rows select t["TABLE_NAME"].ToString()).ToList();

                    conn.Close();
                }
            }

            return tables ?? new List<string>();
        }

        /// <summary>
        /// Fetches the table column metadata.
        /// </summary>
        /// <param name="connectionString">The connection string.</param>
        /// <param name="tableName">Name of the table.</param>
        /// <returns>A list containing each column and it's metadata.</returns>
        public IList<ColumnMetadata> FetchTableColumnMetadata(string connectionString, string tableName)
        {
            var columnMetadatas = new List<ColumnMetadata>();
            var connection = ConnectionFactory.GenerateConnection(DatabaseTypeForHelper, connectionString);

            var conn = connection as OleDbConnection;
            if (conn != null)
            {
                using (conn)
                {
                    conn.Open();

                    var tempCommand = new OleDbCommand(string.Format("SELECT * FROM {0}", tableName), conn);
                    var tempReader = tempCommand.ExecuteReader(CommandBehavior.KeyInfo);
                    if (tempReader != null)
                    {
                        var schemaInfo = tempReader.GetSchemaTable();

                        if (schemaInfo != null)
                        {
                            foreach (DataRow rowColumn in schemaInfo.Rows)
                            {
                                var newMetadata = new ColumnMetadata
                                    {
                                        ColumnName = (string)rowColumn["ColumnName"],
                                        DataType = (Type)rowColumn["DataType"],
                                        IsAutoIncrementField = (bool)rowColumn["IsAutoIncrement"],
                                        IsKeyField = (bool)rowColumn["IsKey"],
                                        NullableField = (bool)rowColumn["AllowDBNull"],
                                        ProviderDataType = (int)rowColumn["ProviderType"],
                                        ColumnOrdinal = (int)rowColumn["ColumnOrdinal"],
                                        ColumnSize = (int)rowColumn["ColumnSize"],
                                        NumericPrecision = (short)rowColumn["NumericPrecision"],
                                        NumericScale = (short)rowColumn["NumericScale"],
                                        IsLong = (bool)rowColumn["IsLong"],
                                        IsReadOnly = (bool)rowColumn["IsReadOnly"],
                                        IsRowVersion = (bool)rowColumn["IsRowVersion"],
                                        IsUnique = (bool)rowColumn["IsUnique"],
                                        BaseSchemaName = rowColumn["BaseSchemaName"],  //////////////////
                                        BaseCatalogName = rowColumn["BaseCatalogName"] == DBNull.Value ? null : (string)rowColumn["BaseCatalogName"],
                                        BaseTableName = rowColumn["BaseTableName"] == DBNull.Value ? null : (string)rowColumn["BaseTableName"],
                                        BaseColumnName = rowColumn["BaseColumnName"] == DBNull.Value ? null : (string)rowColumn["BaseColumnName"]
                                    };

                                columnMetadatas.Add(newMetadata);
                            }
                        }
                        else
                        {
                            Debug.WriteLine("Could not find schema information table.");
                        }
                    }
                    else
                    {
                        Debug.WriteLine("Could not create Data Reader to fetch Ms SQL schema.");
                    }

                    conn.Close();
                }
            }

            return columnMetadatas;
        }
    }
}
